program test
   use ipsfc
   use omp_lib

   implicit none

   integer :: m, n, p=1, i, j, mb1, nb1, mb2, nb2
   integer :: A11, A12, A21, A22
   double precision, dimension(:,:), allocatable :: A
   logical :: ok, cut
   character(4), dimension(6) :: formats = (/'CM', 'CCRB', 'CRRB', 'RCRB', 'RRRB', 'RM'/)
   integer :: f_in, f_out
   integer :: mb1_, nb1_, mb2_, nb2_
   double precision :: T
   integer :: rep, nrep

   mb1 = 64
   nb1 = 64
   mb2 = 64
   nb2 = 64
   nrep = 3

!$omp parallel 
!$omp master
   p = omp_get_num_threads()
!$omp end master
!$omp end parallel

   do f_in = 1, 6
      do f_out = 1, 6
         if( f_in == f_out ) cycle
         do m = (500/mb1)*mb1, (700/mb1)*mb1, mb1
            do n = (500/nb1)*nb1, (700/nb1)*nb1, nb1
               do rep = 1, nrep
                  allocate(A(m,n))
                  call generate_CM(m, n, A(1,1), m)
                  if( formats(f_in) == 'CM' .or. formats(f_in) == 'RM' )then
                     mb1_ = mb1
                     nb1_ = nb1
                  else
                     mb1_ = mb1
                     nb1_ = nb1
                  end if
                  if( formats(f_out) == 'CM' .or. formats(f_out) == 'RM' )then
                     mb2_ = mb1
                     nb2_ = nb1
                  else
                     mb2_ = mb2
                     nb2_ = nb2
                  end if
                  if( formats(f_in) /= 'CM' )then
                     call convert_storage_format(A, A11, A12, A21, A22, m, n, 'CM', mb1_, nb1_, formats(f_in), mb1_, nb1_)
                  end if
                  T = gettime()
                  call convert_storage_format(A, A11, A12, A21, A22, m, n, formats(f_in), mb1_, nb1_, formats(f_out), mb2_, nb2_)
                  T = gettime() - T
                  if( formats(f_out) == 'CM' .or. formats(f_out) == 'RM' )then
                     mb2_ = 1
                     nb2_ = 1
                  else
                     mb2_ = mb2
                     nb2_ = nb2
                  end if
                  ok = matrix_verify(A, A11, A12, A21, A22, formats(f_out), m, n, mb2_, nb2_)
                  !ok = .true.
                  cut = (mod(m,mb1)/=0 .or. mod(n,nb1)/=0 .or. mod(m,mb2)/=0 .or. mod(n,nb2)/=0)
                  write(*,'(3i5,1i3,2i5,1i3,2i5,1f10.6,1e10.3,2i3)') &
                     p, m, n, f_in, mb1, nb1, f_out, mb2, nb2, T, T/(m*n), ltoi(ok), ltoi(cut)
                  deallocate(A)
               end do
            end do
         end do
      end do
   end do

contains

integer function ltoi(l) result(i)
   implicit none
   logical :: l
   intent(in) :: l
   if( l )then
      i = 1
   else
      i = 0
   end if
end function ltoi

logical function matrix_verify(A, A11, A12, A21, A22, informat, m, n, mb, nb) result(ok)
   implicit none
   double precision, dimension(0:m*n-1) :: A
   integer :: A11, A12, A21, A22, m, n, mb, nb
   character(*) :: informat
   intent(in) :: A, A11, A12, A21, A22, m, n, mb, nb, informat
   external :: CM, CCRB, CRRB, RCRB, RRRB, RM
   integer :: CM, CCRB, CRRB, RCRB, RRRB, RM
   ok = .true.
   jloop: do j = 0, n-1
      do i = 0, m-1
         if( informat == 'CM' ) ok = ok .and. element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map_CM)
         if( informat == 'CCRB' ) ok = ok .and. element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map_CCRB)
         if( informat == 'CRRB' ) ok = ok .and. element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map_CRRB)
         if( informat == 'RCRB' ) ok = ok .and. element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map_RCRB)
         if( informat == 'RRRB' ) ok = ok .and. element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map_RRRB)
         if( informat == 'RM' ) ok = ok .and. element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map_RM)
         if( .not. ok ) exit jloop
      end do
   end do jloop
end function matrix_verify

logical function element_verify(A, A11, A12, A21, A22, m, n, mb, nb, i, j, map) result(ok)
   implicit none
   double precision, dimension(0:m*n-1) :: A
   integer :: A11, A12, A21, A22, m, n, mb, nb, i, j
   interface
      integer function map(m, n, mb, nb, i, j)
         implicit none
         integer :: m, n, mb, nb, i, j
         intent(in) :: m, n, mb, nb, i, j
      end function map
   end interface
   intent(in) :: A, A11, A12, A21, A22, m, n, mb, nb, i, j
   integer :: m1, n1, k
   double precision :: expected, found
   expected = 1 + i + j*m
   m1 = (m/mb)*mb
   n1 = (n/nb)*nb
   if( i < m1 .and. j < n1 )then
      k = map(m, n, mb, nb, i, j)
      found = A(A11+k)
   elseif( i < m1 .and. j >= n1 )then
      k = map(m, n-n1, mb, min(nb,n-n1), i, j-n1)
      found = A(A12+k)
   elseif( i >= m1 .and. j < n1 )then
      k = map(m-m1, n, min(mb,m-m1), nb, i-m1, j)
      found = A(A21+k)
   elseif( i >= m1 .and. j >= n1 )then
      k = map(m-m1, n-n1, min(mb,m-m1), min(nb,n-n1), i-m1, j-n1)
      found = A(A22+k)
   end if
   ok = (expected == found)
end function element_verify

end program test
